import numpy as np
import scipy 
from scipy.fftpack import fft2,ifft2,dst,idst
import pandas as pd
from views_transformation_library import utilities


def get_fourier_lag(df, dimensionality, use_stride_tricks=True):

    """
    get_fourier_lag
    
    Driver function for obtaining spatial lags by using Fourier transforms to perform a 
    fast solution of Poisson's equation
    
    Arguments:
    
    df: dataframe containing one or more features to be lagged
    
    dimensionality: dimensionality of Fourier transforms to be used, which controls the 
    distance-weighting of the spatial lag. This can be 
    
    - 2, resulting in a lagging proportional to -ln(d)
    - 3, resulting in a lagging proportional to 1/d
    
    use_stride_tricks: boolean, controlling whether stride tricks is used for tensor
    transforms
    
    """

    if dimensionality == 2:
        transformer = fft_2D
    
    elif dimensionality == 3:
        transformer = fft_3D
    
    else:
        print('Dimensionality ', dimensionality, ' not supported.')
        return None
         
    pgids, pgid_to_longlat, longlat_to_pgid, pgid_to_index,\
    index_to_pgid, ncells, power = utilities._map_pgids_2d(df)
    
    times, time_to_index, index_to_time = utilities._map_times(df)
    
    features = utilities._map_features(df)
    
    tensor4d = utilities._build_4d_tensor(
                                          df,
                                          pgids,
                                          pgid_to_index,
                                          times,
                                          time_to_index,
                                          ncells,
                                          ncells,
                                          pgid_to_longlat,
                                          features,
                                          use_stride_tricks
                                          )
    
    flags = transformer(tensor4d, times, features, time_to_index)
    
    df_flags = flags_to_df(
                           df,
                           flags,
                           times,
                           time_to_index,
                           pgids,
                           features,
                           ncells,
                           ncells,
                           longlat_to_pgid
                           )
    
    return df_flags
    
    
def fft_2D(tensor4d, times, features, time_to_index):

    """

    fft_2D
    
    Function which uses a 2D Fourier transform to solve Poisson's equation in 2 spatial
    dimensions.
    
    This results in distance weighting of the lagged variable of -ln(d)
    
    
    """

    # small value to use instead of zero for zero spatial frequencies

    SMALL = 1e-3

    dim0 = np.shape(tensor4d)[0]
    transformed = np.empty_like(tensor4d)
    
    freqsj = freqsk = None
    rangej = rangek = dim0
    
    for ifeature, feature in enumerate(features):
        for time in times:

            itime = time_to_index[time]
            
            rhomax = np.max(tensor4d[:, :, itime, ifeature])

            # perform 2D FFT on time slice from tensor

            rhojkhat = np.fft.fft2(tensor4d[:, :, itime, ifeature], norm='ortho')

            if freqsj is None:

                # get spatial frequencies

                freqsj = 2.*np.pi*np.fft.fftfreq(rhojkhat.shape[0])
                freqsk = 2.*np.pi*np.fft.fftfreq(rhojkhat.shape[1])
       
                freqsj[abs(freqsj)<SMALL]=SMALL
                freqsk[abs(freqsk)<SMALL]=SMALL

                # make tensor with values 1/(j**2+k**2)

                divisor=np.ones((rangej,rangek))
       
                for ij in range(rangej):
                    ifreqj = freqsj[ij]
                    ifreqj2 = ifreqj*ifreqj
                    for ik in range(rangek):
                        ifreqk = freqsk[ik]
                        ifreqk2 = ifreqk*ifreqk
                        divisor[ij, ik] /= (ifreqj2+ifreqk2)

            # generate solution to Poisson's equation on slice

            phijkhat = np.multiply(rhojkhat, divisor)
            
            # take real part of inverse transform
            
            phijk = np.real(np.fft.ifft2(phijkhat, norm='ortho'))
            
            # subtract zero level
            
            phijk -= np.min(phijk)
            
            # rescale lagged variable to have same scale as original feature
            
            phimax = np.max(phijk)

            phimax = np.max([phimax, 1.0])
            
            transformed[:, :, itime, ifeature] = phijk*rhomax/phimax

    return transformed


def fft_3D(tensor4d, times, features, time_to_index):

    """
    fft_3D
    
    Function which uses a 3D Fourier transform to solve Poisson's equation in 3 spatial
    dimensions.
    
    This results in distance weighting of the lagged variable of 1/d
    
    This is done by embedding the flat array from the original feature as the midplane of 
    a cube, so that the problem effectively becomes 3-dimensional.
    
    To save memory and time, the number of Fourier modes used in the extra dimension is 
    reduced by a factor LDIVISOR. Experimentation shows that adequate resolution can be
    achieved with LDIVISOR set to 4
    
    
    """

    # factor by which to reduce number of Fourier modes in extra dimension

    LDIVISOR = 4
    
    # small value to use instead of zero for zero spatial frequencies
    
    SMALL = 1e-3
    
    dim0 = np.shape(tensor4d)[0]
    
    transformed = np.empty_like(tensor4d)

    freqsj = freqsk = freqsl = None
    rangej = rangek = dim0
    rangel = int(dim0/LDIVISOR)
    
    for ifeature, feature in enumerate(features):
        for time in times:
          
            itime = time_to_index[time]
            
            rhomax = np.max(tensor4d[:, :, itime, ifeature])
            
            # embed time slice from tensor into a 3D tensor

            rhojkl = np.zeros((rangej, rangek, rangel))
        
            icentrej = icentrek = int(rangej/2)
            icentrel = int(rangel/2)
        
            rhojkl[:, :, icentrel] = tensor4d[:, :, itime, ifeature]
            
            # perform N-dimensional FFT on embedded time slice

            rhojklhat = np.fft.fftn(rhojkl, norm='ortho')

            if freqsj is None:

                # get spatial frequencies

                freqsj = 2.*np.pi*np.fft.fftfreq(rhojklhat.shape[0])
                freqsk = 2.*np.pi*np.fft.fftfreq(rhojklhat.shape[1])
                freqsl = 2.*np.pi*np.fft.fftfreq(rhojklhat.shape[2])
       
                freqsj[abs(freqsj)<SMALL] = SMALL
                freqsk[abs(freqsk)<SMALL] = SMALL
                freqsl[abs(freqsl)<SMALL] = SMALL

                # make tensor with values 1/(j**2+k**2+l**2)

                divisor = np.ones((rangej, rangek, rangel))
       
                for ij in range(rangej):
                    ifreqj = freqsj[ij]
                    ifreqj2 = ifreqj*ifreqj
                    for ik in range(rangek):
                        ifreqk = freqsk[ik]
                        ifreqk2 = ifreqk*ifreqk
                        for il in range(rangel):
                            ifreql = freqsl[il]
                            ifreql2 = ifreql*ifreql
            
                            divisor[ij, ik, il] /= (ifreqj2+ifreqk2+ifreql2)

            # generate solution to Poisson's equation on embedded slice

            phijklhat = np.multiply(rhojklhat, divisor)
            
            # take real part of inverse transform
            
            phijkl = np.real(np.fft.ifftn(phijklhat, norm='ortho'))
            
            # subtract zero level
            
            phijkl -= np.min(phijkl)
            
            # rescale lagged variable to have same scale as original feature
            
            phimax = np.max(phijkl)

            phimax = np.max([phimax, 1.0])

            transformed[:, :, itime, ifeature] = phijkl[:, :, icentrel]*rhomax/phimax
        
    return transformed


def flags_to_df(
    df,
    flags,
    times,
    time_to_index,
    pgids,
    features,
    longrange,
    latrange,
    longlat_to_pgid
):

    final = np.empty((len(times)*len(pgids),len(features)))

    # convert 4d spatial lag tensor to flat table

    ipgid = 0
    npg = len(pgids)
    pgids_for_index = []

    for ilong in range(1, longrange+2):
        for ilat in range(1, latrange+2):
            try:
                pgid = longlat_to_pgid[(ilong, ilat)]
                pgids_for_index.append(pgid)
                for time in times:
                    tindex = time_to_index[time]
                    indx = tindex*npg+ipgid
                    final[indx, :] = flags[ilong, ilat, tindex, :]

                ipgid += 1
            except:
                pass
            
    index_names = df.index.names
        
    df_index = pd.MultiIndex.from_product([list(times), pgids_for_index], names=index_names)
            
    colnames = ['flag_'+feature for feature in features]
    df_flags = pd.DataFrame(data=final, columns=colnames, index=df_index)

    df_flags = df_flags.sort_index()

    return df_flags



import numpy as np
import pandas as pd

from scipy import ndimage

from views_transformation_library import utilities


def get_splag4d(
    df,
    use_stride_tricks=True,
    kernel_inner=1, kernel_width=1, kernel_power=0, norm_kernel=0

):

    """
    splag4d created 19/03/2021 by Jim Dale
    
    Performs spatial lags on a dataframe by transforming from flat format to 4d tensor 
    with dimensions longitude x latitude x time x features.
    
    Spatial lagging can then be done as a 2d convolution on long-lat slices using
    scipy convolution algorithms.
    
    Arguments:
    
    df:                a dataframe of series to be splagged
    
    use_stride_tricks: boolean, decide to use stride_tricks or not (optional, 
                       defaults to True)
   
    kernel_inner:      inner border of convolution region (set to 1 to exclude central
                       cell)

    kernel_width:      width in cells of kernel, so outer radius of kernel =
                       kernel_inner + kernel_width

    kernel_power:      weight values of cells by (distance from centre of kernel)**
                       (-kernel_power) - set to zero for no distance weighting

    norm_kernel:       set to 1 to normalise kernel weights
                       
    Returns:
    
    A df of all lagged columns from input df
    
    """

    arg_string = str(kernel_inner) + '_' + str(kernel_width) + '_' + str(kernel_power) + '_' + str(norm_kernel)

    df = df.fillna(0.0)
    if not df.index.is_monotonic:
        df = df.sort_index()
 
    weights = build_kernel_weights(kernel_inner, kernel_width, kernel_power, norm_kernel)
     
    times, time_to_index, index_to_time = utilities._map_times(df)

    features = utilities._map_features(df)

    pgids, pgid_to_longlat, longlat_to_pgid, pgid_to_index, index_to_pgid, ncells, power =\
                                                          utilities._map_pgids_2d(df)
    
    tensor4d = utilities._build_4d_tensor(
                                          df,
                                          pgids,
                                          pgid_to_index,
                                          times,
                                          time_to_index,
                                          ncells,
                                          ncells,
                                          pgid_to_longlat,
                                          features,
                                          use_stride_tricks
                                          )

    splags = get_splags(tensor4d, ncells, ncells, times, features, time_to_index, weights)

    df_splags = splags_to_df(
                             df,
                             splags,
                             times,
                             time_to_index,
                             pgids,
                             features,
                             ncells,
                             ncells,
                             longlat_to_pgid,
                             arg_string
                             )

    return df_splags


def build_kernel_weights(kernel_inner, kernel_width, kernel_power, norm_kernel):
    
    kernel_inner = int(kernel_inner)
    kernel_width = int(kernel_width)
    
    kernel_size = int(2*(kernel_inner+kernel_width)) - 1
    weights = np.ones((kernel_size, kernel_size))
    
    kernel_centre = int((kernel_size+1)/2) - 1
    
    for ix in range(kernel_size):
        dx = ix-kernel_centre
        for iy in range(kernel_size):
            dy = iy-kernel_centre
            if (abs(dx) < kernel_inner) and (abs(dy) < kernel_inner):
                weights[ix, iy] = 0.0
            else:
                rxy = np.sqrt(dx*dx+dy*dy)
                weights[ix, iy] /= rxy**kernel_power
               
    if norm_kernel:
        weights /= np.sum(weights)
       
    return weights


def get_splags(
    tensor4d,
    longrange,
    latrange,
    times,
    features,
    time_to_index,
    weights
):

    splags = np.empty((longrange, latrange, len(times), len(features)))
    
    # use scipy convolution function to do 2d convolution on tensor slices
    
    for time in times:
        tindex = time_to_index[time]
        
        for ifeature in range(len(features)):
            raw = tensor4d[:, :, tindex, ifeature]
            splags[:, :, tindex, ifeature] = ndimage.convolve(raw, weights, mode='constant', cval=0.0)

    return splags


def splags_to_df(
    df,
    splags,
    times,
    time_to_index,
    pgids,
    features,
    longrange,
    latrange,
    longlat_to_pgid,
    argstring
):

    final=np.empty((len(times)*len(pgids),len(features)))

    # convert 4d spatial lag tensor to flat table

    ipgid = 0
    npg = len(pgids)
    pgids_for_index = []

    for ilong in range(1, longrange+2):
        for ilat in range(1, latrange+2):
            try:
                pgid = longlat_to_pgid[(ilong, ilat)]
                pgids_for_index.append(pgid)
                for time in times:
                    tindex = time_to_index[time]
                    indx = tindex*npg+ipgid
                    final[indx, :] = splags[ilong, ilat, tindex, :]

                ipgid += 1
            except:
                pass
            
    index_names = df.index.names
        
    df_index = pd.MultiIndex.from_product([list(times), pgids_for_index], names=index_names)
        
#    colnames=['splag_'+argstring+'_'+feature for feature in features]

    colnames = df.columns

    df_splags = pd.DataFrame(data=final, columns=colnames, index=df_index)

    df_splags = df_splags.sort_index()

    return df_splags



import pandas as pd
import numpy as np
from views_transformation_library import utilities


def get_tree_lag(df, thetacrit, weight_functions, sigma, use_stride_tricks):
    
    """
    get_tree_lag
    
    Driver function for computing temporal-tree-lagged features
    
    Arguments are:
    
    df: dataframe containing one or more features to be lagged
    thetacrit: parameter controlling how aggressively nodes in the past are aggregated
    weight_functions: list containing one or more of the following strings:
    
    - 'uniform': weights for all nodes are unity. Unlikely to be meaningful but provided for completeness
    - 'oneovert': weights for nodes are 1/(tnow-tnode)
    - 'expon': weights for nodes are exp(-(tnow-tnode)/sigma)
    - 'ramp': weights for nodes are 1-(tnow-tnode)/sigma for (tnow-tnode)<sigma, 0 otherwise
    - 'sigmoid': weights are 1./(1+np.exp(-lag)) where lag=(mid-tnow+5*sigma5)/sigma5 and sigma5=sigma/5
    
    sigma: parameter in time units used by df controlling width of expon, ramp and sigmoid functions
    
    """
    
    df = df.fillna(0.0)
    if not df.index.is_monotonic:
        df = df.sort_index()
        
    if type(weight_functions) != 'list':
        weight_functions = [weight_functions, ]
    
    tree = TemporalTree()

    tree.build_tree(df)
    
    tree.stock_initial(df, use_stride_tricks)
    
    df_treelags = tree.tree_lag(df, thetacrit, weight_functions, sigma)
    
    return df_treelags


class TemporalNode():
    """
    TemporalNode() class
    Defines nodes for insertion in TemporalTree() class
    
     Nodes have:
    - nodeid: a unique id
    - level: integer indicating which level of the tree the node belongs to, 0 is leaf level
    - parent: parent node
    - start, end: node boundaries
    - sibling: id of node's sibling
    - predecessor: id of immediately preceding node at same tree level
    - ispast: flag indicating whether node is past or future compared to its sibling 
    - isleaf: a flag indicating whether a leaf node
    - nleaf: the number of leaf nodes contained within the node's boundaries
    - children: list of node's children
    - features: dict of features loaded into the node by the .stock() method
    
    """    
    def __init__(self, nodeid, level, start, end, parent, sibling, predecessor, ispast, isleaf):
        self.nodeid = nodeid
        self.level = level
        self.start = start
        self.end = end
        self.parent = parent
        self.sibling = sibling
        self.predecessor = predecessor
        self.ispast = ispast
        self.isleaf = isleaf
        self.children = []
        self.nleaf = 0
        self.features = {}
        
    def __repr__(self):
        return f""" nodeid: {self.nodeid}\n level: {self.level}\n parent: {self.parent}\n start: {self.start}
        end: {self.end}\n bottom: {self.bottom}\n isleaf: {self.isleaf}\n children: {self.children}
        predecessor: {self.predecessor}\n sibling: {self.sibling}\n"""


class TemporalTree():
    """
    TemporalTree() class
    """  
    def __init__(self):
        self.gridsize = 0
        self.nodes = []
        self.pgids = None
        self.maxlevel = -1
        self.stocked = False
        self.stocked_until = -1
        self.weight_fn = None
        self.timesequence = []
        self.npad = 0
        self.times = []
        self.time_to_index = {}
        self.index_to_time = {}
        self.pgid_to_index = {}
        self.index_to_pgid = {}
        self.features = {}
        self.weight_functions = {}
        
        self.make_weight_functions()
    
    def build_tree(self, df):
       
        """
        build_tree

        This function builds the tree from the leaf nodes downwards.

        The total number of leaf nodes must be an integer power of two, so the list of
        times is padded with dummy nodes placed before the earliest time the tree is
        built on.

        Once leaf nodes are populated, work down through the levels. Assign parents,
        siblings and predecessors

        """

        self.times, self.time_to_index, self.index_to_time = utilities._map_times(df)
        
        tstart = self.times[0]
        tend = self.times[-1]
        
        nseq_initial = tend-tstart
        log2_nseq_initial = np.log2(nseq_initial)
        nseq = int(2**(1+int(log2_nseq_initial)))
        npad = nseq-nseq_initial
        
        self.timesequence = [t for t in range(tstart-npad, tend+1)]
        self.npad = npad
        
        nodestodo = []
        nnodes = 0
        level = 0
        maxlevel = 0
        parent = -1
        sibling = -1
        predecessor = -1
        ispast = False
        isleaf = False
        node = TemporalNode(nnodes, level, tstart-npad, tend, parent, sibling, predecessor, ispast, isleaf)
        self.nodes.append(node)
        nodestodo.append(node)
        
        while len(nodestodo) > 0:
            splitnode = nodestodo.pop()
            
            if (splitnode.end-splitnode.start) > 1:
                nnodes += 1
                if splitnode.start >= 0:
                    mid = int((splitnode.start+splitnode.end+1)/2)
                else:
                    mid = int((splitnode.start+splitnode.end)/2)

                level = splitnode.level+1
                maxlevel = max(maxlevel, level)
                parent = splitnode.nodeid
                sibling = None
                predecessor = None
                ispast = True
                if (mid-splitnode.start) > 1:
                    isleaf = False
                else:
                    isleaf = True
                pastnode = TemporalNode(nnodes, level, splitnode.start, mid, parent, sibling, predecessor, ispast,
                                        isleaf)

                self.nodes.append(pastnode)

                nodestodo.append(pastnode)

                nnodes += 1
                ispast = False
                futurenode = TemporalNode(nnodes, level, mid, splitnode.end, parent, sibling, predecessor, ispast,
                                          isleaf)

                self.nodes.append(futurenode)

                nodestodo.append(futurenode)

                pastnode.sibling = futurenode.nodeid

                futurenode.sibling = pastnode.nodeid

                futurenode.predecessor = pastnode.nodeid

                splitnode.children = [pastnode.nodeid, futurenode.nodeid]
                
        self.maxlevel = maxlevel
        
        for node in self.nodes:
            while node.predecessor is None:
                if node.start <= (self.timesequence[0]+self.npad):
                    node.predecessor = -1
                else:
                    level = node.level
                    climb = True
                    climbnode = self.nodes[node.parent]
                    while climb:
                        if climbnode.ispast:
                            climbnode = self.nodes[climbnode.parent]
                        else:
                            descendnode = self.nodes[climbnode.sibling]
                            climb = False
                    descendlevel = descendnode.level

                    while descendnode.level != level:
                        descendnode = self.nodes[descendnode.children[1]]
                    node.predecessor = descendnode.nodeid
        
    def stock_initial(self, df, use_stride_tricks):
    
        """
        stock_initial
        
        This function loads features into the tree.
        
        Data from the input df is first cast to a 3D tensor with dims time x pgid x
        features.
        
        The tensor is used to populate the leaf nodes, and values then just propagate
        down through the tree via the parents.
        
        At present, values of parents are the *sums* of those of their children.
        
        """
    
        if self.stocked:
            print('Tree has already been stocked - aborting')
            raise(Exception)
            
        self.stocked = True

        self.features = utilities._map_features(df)
        
        self.pgids, self.pgid_to_index, self.index_to_pgid = utilities._map_pgids_1d(df)

        npgids = len(self.pgids)

        for node in self.nodes:
            for feature in self.features:
                node.features[feature] = np.zeros(npgids)
                
        if use_stride_tricks:
            tensor3d = utilities._df_to_tensor_strides(df)
        else:
            tensor3d = utilities._df_to_tensor_strides(df)

        for time in self.times:
            itime = self.time_to_index[time]
            for node in self.nodes:
                if node.isleaf and (node.start == time):
                    for ifeature, feature in enumerate(self.features):
                        vals = tensor3d[itime, :, ifeature]
                        node.features[feature] += vals
                        node.nleaf = 1
                        parentid = node.parent
                        while parentid != -1:
                            parent = self.nodes[parentid]
                            parent.features[feature] += vals
                            parent.nleaf += 1
                            parentid = parent.parent
                
        self.stocked_until = self.times[-1]
        
    def walk(self, tnow, thetacrit):
        
        """
        walk

        This function generates the list of nodes any given node will import data from,
        when one lags variables using the tree, as well as weights based on the times
        between nodes' midpoints.

        The arguments are

        - thetacrit: angle used to decide if a candidate node should be added to a target
          node's interaction list, based on the size of the candidate node and
          the time gap between the candidate node and the target node

        """
        
        if (tnow<self.nodes[0].start) or (tnow>self.nodes[0].end):
            print('tnow not in range of times covered by this tree - aborting')
            raise Exception
        if tnow > self.stocked_until:
            print('tree has not been stocked as far as tnow - aborting')
            raise Exception
            
        list_of_nodes=[]

        for node in self.nodes:
            if node.isleaf and (node.start == tnow):
                list_of_nodes.append(node.nodeid)
                if node.predecessor == -1:
                    return list_of_nodes
                notdone = True
                while notdone:
                    if node.ispast:
                        if node.predecessor == -1:
                            notdone = False
                        else:
                            pred = self.nodes[node.predecessor]
                            node = self.nodes[pred.parent]
                            self.split_node(node, list_of_nodes, tnow, thetacrit)

                    else:
                        node = self.nodes[node.sibling]
                        self.split_node(node, list_of_nodes, tnow, thetacrit)
                        node = self.nodes[node.parent]
                        if node.predecessor == -1:
                            notdone = False
                        else:
                            if node.sibling != node.predecessor:
                                node = self.nodes[node.predecessor]
                                self.split_node(node, list_of_nodes, tnow, thetacrit)
                
        return list_of_nodes
    
    def split_node(self, node, list_of_nodes, tnow, thetacrit):
    
        """
        split_node

        Function which decides whether or not to split a given node into its children,
        based on the critical angle and the current time
        """

        nodestocheck = []
        nodestocheck.append(node)
        while len(nodestocheck) > 0:
            node = nodestocheck.pop(0)
            mid = (node.start+node.end)/2.
            width = (node.end-node.start)
            age = tnow-mid
            theta = width/age

            if theta < thetacrit:
                list_of_nodes.append(node.nodeid)
            else:
                if len(node.children) > 0:
                    nodestocheck.append(self.nodes[node.children[0]])
                    nodestocheck.append(self.nodes[node.children[1]])
                else:
                    list_of_nodes.append(node.nodeid)

    def make_weight_functions(self):
        self.weight_functions['uniform'] = self._uniform
        self.weight_functions['oneovert'] = self._oneovert
        self.weight_functions['sigmoid'] = self._sigmoid
        self.weight_functions['expon'] = self._expon
        self.weight_functions['ramp'] = self._ramp
                    
    def _uniform(self, node_list, tnow, sigma=1):
        weights = np.ones(len(node_list))
        return weights

    def _oneovert(self, node_list, tnow, sigma=1):
        weights = np.zeros(len(node_list))
        for i in range(len(node_list)):
            mid = (self.nodes[node_list[i]].start+self.nodes[node_list[i]].end)/2.
            lag = tnow-mid+1.5
            weights[i] = 1./lag
        return weights
    
    def _sigmoid(self, node_list, tnow, sigma=1):
        weights = np.zeros(len(node_list))
        offset = 5.
        sigma /= offset
        for i in range(len(node_list)):
            mid = self.nodes[node_list[i]].start
            lag = (mid-tnow+offset*sigma)/sigma
            weights[i] = 1./(1+np.exp(-lag))

        return weights
    
    def _expon(self, node_list, tnow, sigma=1):
        weights = np.zeros(len(node_list))
        for i in range(len(node_list)):
            mid = (self.nodes[node_list[i]].start+self.nodes[node_list[i]].end)/2.
            lag = tnow-mid
            lag1 = tnow-self.nodes[node_list[i]].start
            
            lag2 = tnow-self.nodes[node_list[i]].end
            
            w = np.exp(-lag/sigma)
            w1 = np.exp(-lag1/sigma)
            w2 = np.exp(-lag2/sigma)
            weights[i] = (8*w1+6*w-w2)/13.

        return weights
    
    def _ramp(self, node_list, tnow, sigma=1):
        weights = np.zeros(len(node_list))
        for i in range(len(node_list)):
            mid = (self.nodes[node_list[i]].start+self.nodes[node_list[i]].end)/2.
            lag = tnow-mid+0.5
            weights[i] = 1.-(lag/sigma)

        weights[weights < 0.0] = 0.0
        return weights
    
    def tree_lag(self, df, thetacrit, weight_functions, sigma):
    
        """
        tree_lag
        
        This function computes features temporally lagged using the tree for every pgid.
        
        The interaction lists and weights, and the features loaded into each node are 
        combined in weighted sums.
        
        The lagged features are then packed into a dataframe.
        
        """
    
        ntimes = len(self.times)
        npgids = len(self.pgids)
        nfeatures = len(self.features)
        nfunctions = len(weight_functions)
    
        tensor3d = np.empty((ntimes, npgids, nfeatures*nfunctions))
    
        for time in self.times:
            itime = self.time_to_index[time]
            list_of_nodes = self.walk(time, thetacrit)
            for ifunction, weight_function in enumerate(weight_functions):
                weights = self.weight_functions[weight_function](list_of_nodes, time, sigma)
                for ifeature, feature in enumerate(self.features):
                    weighted_feature = np.zeros_like(self.nodes[0].features[feature])
                    for nodeid, weight in zip(list_of_nodes, weights):
                        weighted_feature += weight*self.nodes[nodeid].features[feature]
        
                    tensor3d[itime, :, ifeature*nfunctions+ifunction] = weighted_feature
                    
        flat=np.empty((ntimes*npgids,nfeatures*nfunctions))
    
        for ichunk in range(ntimes):
            flat[ichunk*npgids:(ichunk+1)*npgids, :] = tensor3d[ichunk, :, :]

        df_column_names = df.columns

        index_names = df.index.names
        
        df_index = pd.MultiIndex.from_product([self.times, self.pgids], names=index_names)
    
        df_treelags = pd.DataFrame(flat, index=df_index, columns=df_column_names)
    
        return df_treelags



import numpy as np
import xarray
import ingester3
from ingester3.Country import Country
from ingester3.scratch import fetch_data

def _df_to_tensor_no_strides(df):
    '''
    df_to_tensor_no_strides created 13/03/2021 by Jim Dale
    Uses regular array indexing to create a tensorlike from a dataframe.

    dim0 of the tensor corresponds to level 0 of the df multiindex,
    dim1 corresponds to level 1 of the df multiindex, and dim2 corresponds to
    the df's columns.
    '''


    # get shape of dataframe

    dim0,dim1=df.index.levshape

    dim2=df.shape[1]

    # check that df can in principle be tensorised

    if dim0*dim1!=df.shape[0]:
        raise Exception("df cannot be cast to a tensor - dim0 * dim1 != df.shape[0]",dim0,dim1,df.shape[0])


    flat=df.to_numpy()

    tensor3d=np.empty((dim0,dim1,dim2))
    chunksize=dim1
    nchunks=dim0
    for ichunk in range(nchunks):
        tensor3d[ichunk,:,:]=flat[ichunk*chunksize:(ichunk+1)*chunksize,:]

    return tensor3d 

def _df_to_tensor_strides(df):
    '''
    df_to_tensor created 13/03/2021 by Jim Dale
    Uses as_strided from numpy stride_tricks library to create a tensorlike
    from a dataframe.

    dim0 of the tensor corresponds to level 0 of the df multiindex,
    dim1 corresponds to level 1 of the df multiindex, and dim2 corresponds to
    the df's columns.
    '''

    # get shape of dataframe

    dim0,dim1=df.index.levshape

    dim2=df.shape[1]

    # check that df can in principle be tensorised

    if dim0*dim1!=df.shape[0]:
        raise Exception("df cannot be cast to a tensor - dim0 * dim1 != df.shape[0]",
                        dim0,dim1,df.shape[0])

    flat=df.to_numpy()

    # get strides (in bytes) of flat array
    flat_strides=flat.strides

    offset2=flat_strides[1]

    offset1=flat_strides[0]

    # compute stride in bytes along dimension 0
    offset0=dim1*offset1

    # get memory view or copy as a numpy array
    tensor3d=np.lib.stride_tricks.as_strided(flat,shape=(dim0,dim1,dim2),
                                           strides=(offset0,offset1,offset2))

    return tensor3d
    

def _map_times(df):
    '''
    
    map_times
    
    Gets the times from a dataframe, assuming that they are the first index, and returns
    a list of unique times, and dicts which convert between the actual time, and the
    index in the list.

    '''
    # get unique times

    times=np.array(list({idx[0] for idx in df.index.values}))
    times=list(np.sort(times))
    
    # make dicts to transform between times and the index of a time in the list
    
    time_to_index={}
    index_to_time={}
    for i,time in enumerate(times):
        time_to_index[time]=i
        index_to_time[i]=time
    
    return times,time_to_index,index_to_time
    
def _map_features(df):
    '''
    
    map_features
    
    Gets a list of features from a dataframe.

    '''

    features=list(df.columns)
    
    return features
    
def _map_pgids_1d(df):
    '''
    
    map_pgid_1d
    
    Gets the pgids from a dataframe, assuming that they are the second index, and 
    returns a 1d list of unique pgids, together with dicts which convert between 
    the actual pgid and the index in the list.
    
    '''
    
    pgids=np.array(list({idx[1] for idx in df.index.values}))

    pgids=np.sort(pgids)
    
    pgid_to_index={}
    index_to_pgid={}

    for i,pgid in enumerate(pgids):
        pgid_to_index[pgid]=i
        index_to_pgid[i]=pgid

    return pgids,pgid_to_index,index_to_pgid
    
    
def _map_pgids_2d(df):
    '''
	map_pgids_2d
	
	This function builds a 2D map in longitude-latitude from the pgids contained in
	the input dataframe assuming they are the second index column, and creates dicts 
	allowing quick transformation from (long,lat) to pgid and vice versa.
	
	The pgids are embedded and centred in the smallest possible square grid whose side
	is an integer power of 2

	'''
	
    PG_STRIDE=720
	
	# get unique pgids
	
    pgids=np.array(list({idx[1] for idx in df.index.values}))
    pgids=np.sort(pgids)
		
	# convert pgids to longitudes and latitudes
	
    longitudes=pgids%PG_STRIDE
    latitudes=pgids//PG_STRIDE
	
    latmin=np.min(latitudes)
    latmax=np.max(latitudes)
    longmin=np.min(longitudes)
    longmax=np.max(longitudes)

    latrange=latmax-latmin
    longrange=longmax-longmin

	# shift to a set of indices that starts at [0,0]

    latitudes-=latmin
    longitudes-=longmin

    # find smallest possible square grid with side 2^ncells which will fit the pgids

    latmin=np.min(latitudes)
    latmax=np.max(latitudes)
    longmin=np.min(longitudes)
    longmax=np.max(longitudes)

    maxsize=np.max((longrange,latrange))
    power=1+int(np.log2(maxsize))

    ncells=2**power

    # centre the pgids

    inudgelong=int((ncells-longmax)/2)
    inudgelat=int((ncells-latmax)/2)

    longitudes+=inudgelong
    latitudes+=inudgelat

	# make dicts to transform between pgids and (long,lat) coordinate

    pgid_to_longlat={}
    longlat_to_pgid={}
    
    pgid_to_index={}
    index_to_pgid={}

    for i,pgid in enumerate(pgids):
        pgid_to_longlat[pgid]=(longitudes[i],latitudes[i])
        longlat_to_pgid[(longitudes[i],latitudes[i])]=pgid
        pgid_to_index[pgid]=i
            
    return pgids,pgid_to_longlat,longlat_to_pgid,pgid_to_index,index_to_pgid,ncells,power
    
def _build_4d_tensor(
    df,
    pgids,
    pgid_to_index,
    times,
    time_to_index,
    longrange,
    latrange,
    pgid_to_longlat,
    features,
    use_stride_tricks
):
    '''
    
    build_4d_tensor
    
    This function accepts an input dataframe and the dicts generated by the map_times,
    map_features and map_pgids_2d functions, and returns a 4d latitude x longitude x
    time x feature tensor.

    
    '''
    
    # convert flat data from df into time x pgid x feature tensor
    
    if(use_stride_tricks):
        tensor3d=_df_to_tensor_strides(df)
    else:
        tensor3d=_df_to_tensor_strides(df)

    # convert 3d tensor into longitude x latitude x time x feature tensor

    tensor4d=np.zeros((longrange,latrange,len(times),len(features)))
    for pgid in pgids:

        pgindex=pgid_to_index[pgid]
        for time in times:
        
            tindex=time_to_index[time]
            ilong=pgid_to_longlat[pgid][0]
            ilat=pgid_to_longlat[pgid][1]
            
            tensor4d[ilong,ilat,tindex,:]=tensor3d[tindex,pgindex,:]
                 
    return tensor4d    
    
def _df_to_datacube(df,use_stride_tricks):

    '''

    df_to_datacube
    
    This function streamlines the pgid, time and feature mapping, together with the 4d
    tensor build, so that a df is accepted and a 4d tensor is produced to enable quick
    plotting.

    '''

    pgids,pgid_to_longlat,longlat_to_pgid,pgid_to_index,index_to_pgid,ncells,power=_map_pgids_2d(df)

    times,time_to_index,index_to_time=_map_times(df)

    features=_map_features(df)
    
    tensor=_build_4d_tensor(
                           df,
                           pgids,
                           pgid_to_index,
                           times,
                           time_to_index,
                           ncells,
                           ncells,
                           pgid_to_longlat,
                           features,
                           use_stride_tricks)
    
    return tensor
    
def _get_country_neighbours_tensor():

    '''
    
    get_country_neighbours_tensor
    
    Fetches the country_country_month_expanded table and transforms it to a tensor with
    dimensions month x country x country, where (month_i, country_a, country_b)=1 
    indicates that country_a and country_b are neighbours in month_i. 
    
    Note that (month_i, country_a, country_a)=0, and that countries which never have any 
    neighbours do not appear in the tensor.
    
    '''

    columns = ['month_id', 'a_id', 'b_id']

    df_neighbs = fetch_data(loa_table='country_country_month_expanded', columns=columns)

    df_neighbs=df_neighbs[columns].sort_values(
                                               ["month_id", "a_id","b_id"], 
                                               ascending = (True, True, True)
                                               )

    neighb_month_ids=np.sort(df_neighbs['month_id'].unique())

    neighb_country_ids=np.sort(df_neighbs['b_id'].unique())

    nmonths=len(neighb_month_ids)
    
    ncountries=len(neighb_country_ids)

    neighb_data=np.zeros((nmonths,ncountries,ncountries))
    
    neighb_tensor_dims=['month', 'country', 'neighbours']
    neighb_tensor_coords={}
    neighb_tensor_coords['month']=neighb_month_ids
    neighb_tensor_coords['country']=neighb_country_ids
    neighb_tensor_coords['neighbours']=neighb_country_ids
    
    
    # build dicts to tranform between month and index and country and index in the input
    # dataframe
    
    month_to_index={}
    for imonth,month in enumerate(neighb_month_ids):
        month_to_index[month]=imonth
        
    country_to_index={}
    for icountry,country in enumerate(neighb_country_ids):
        country_to_index[country]=icountry

    neighbours_table=df_neighbs.to_numpy()

    # assemble tensor

    for i in range(np.shape(neighbours_table)[0]):
        month=neighbours_table[i][0]
        imonth=month_to_index[month]
        
        country=neighbours_table[i][1]
        icountry=country_to_index[country]
        
        neighb=neighbours_table[i][2]
        ineighb=country_to_index[neighb]
                 
        neighb_data[imonth,icountry,ineighb]=1

    neighb_tensor=xarray.DataArray(data=neighb_data, coords=neighb_tensor_coords, dims=neighb_tensor_dims)

    return neighb_tensor
        
def _get_country_distances(country_ids,country_id_to_index):

    '''
    
    get_distances
    
    Fetched distances between countries using the Country class and stores them in a
    country x country tensor.
    
    '''
    
    ncountries=len(country_ids)
    
    distances=np.zeros((ncountries,ncountries))

    for country_idi in country_ids:
        country_indexi=country_id_to_index[country_idi]
            
        countryi=Country(country_idi)
                          
        lati=countryi.lat
        longi=countryi.lon
                       
        for country_idj in country_ids[country_indexi:]:
            country_indexj=country_id_to_index[country_idj]
            
            countryj=Country(country_idj)
                
            latj=countryj.lat
            longj=countryj.lon
                
            distances[country_indexi,country_indexj]= \
            distances[country_indexj,country_indexi]= \
            np.sqrt((lati-latj)**2.+(longi-longj)**2.)
        
    return distances    
    



import numpy as np
import pandas as pd
from views_transformation_library import utilities


def tlags3d(
    df,
    use_stride_tricks=True,
    lags=[1, ]
):

    """
    tlag3d created 22/03/2021 by Jim Dale
    
    Performs temporal lags on a dataframe by transforming from flat format to 3d tensor 
    with dimensions time x pgid x features.
    
    Accepts arbitrary numbers of columns in input df, and arbitrary list of lags.
    
    Arguments:
    
    df:                a dataframe of series to be timelagged
    
    use_stride_tricks: boolean, decide to use stride_tricks or not (optional, 
                       defaults to True)
                       
    lags:              a list of lags to be applied (optional, defaults to [1,])
    
    Returns:
    
    A dataframe with all lags of all columns in input df
    
    """

#    df = df.fillna(0.0)
    df = df.sort_index()
    df_index = df.index

    times, time_to_index, index_to_time = utilities._map_times(df)

    features = utilities._map_features(df)

    pgids, pgid_to_index, index_to_pgid = utilities._map_pgids_1d(df)

    if use_stride_tricks:
        tensor3d = utilities._df_to_tensor_strides(df)
    else:
        tensor3d = utilities._df_to_tensor_no_strides(df)

    tlags = get_tlags(times, pgids, features, time_to_index, tensor3d, lags)

    df_tlags = tlags_to_df_strides(tlags, times, time_to_index, pgids, features, lags, df_index)

    return df_tlags


def get_tlags(
    times,
    pgids,
    features,
    time_to_index,
    tensor3d,
    lags
):

    tlags = np.zeros((len(times), len(pgids), len(features)*len(lags)))

    for ifeature in range(len(features)):
        for ilag, lag in enumerate(lags):
            for time in times:
                tindex = time_to_index[time]
                try:
                    tlags[tindex, :, ifeature*len(lags)+ilag] = tensor3d[tindex-lag, :, ifeature]
                except:
                    pass  

    return tlags


def tlags_to_df_no_strides(
    tensor3d,
    tlags,
    times,
    time_to_index,
    pgids,
    features,
    lags,
    df_index
):

    dim0, dim1, dim2 = len(times), len(pgids), len(features)*len(lags)
    
    flat = np.empty((dim0*dim1, dim2))
    
    for ichunk in range(dim0):
        flat[ichunk*dim1:(ichunk+1)*dim1, :] = tensor3d[ichunk, :, :]
       
    df_column_names = df.columns

    df_tlags = pd.DataFrame(flat, index=df_index, columns=df_column_names)
    
    return df_tlags    
    
    
def tlags_to_df_strides(
    tlags,
    times,
    time_to_index,
    pgids,
    features,
    lags,
    df_index
):

    dim0, dim1, dim2 = len(times), len(pgids), len(features)*len(lags)
    
    flat = np.empty((dim0*dim1, dim2))
    
    tlags_strides=tlags.strides

    offset2 = tlags_strides[2]

    offset1 = tlags_strides[1]
    
    offset0 = tlags_strides[0]
    
    flat = np.lib.stride_tricks.as_strided(tlags, shape=(dim0*dim1, dim2),
                                           strides=(offset1, offset2))

    df_column_names = df.columns

    df_tlags = pd.DataFrame(flat, index=df_index, columns=df_column_names)
    
    return df_tlags    
    


import numpy as np
import pandas as pd

from views_transformation_library import utilities


def get_temporal_entropy(
        df,
        window,
        offset=0.

):
    """"
    get_temporal_entropy created 04/03/2022 by Jim Dale

    Computed entropy along the time axis within a window of length specified by 'window'.

    The entropy of a feature x over a window of length w is

    sum_(i=1,w) (x_i/X)log_2(x_i/X) where X = sum_(i=1,w) (x_i)

    Arguments:

    df:                a dataframe of series to be splagged

    window:            integer size of window

    offset:            datasets containing mostly zeros will return
                       NaNs or Infs for entropy most or all of the time.
                       Since this is unlikely to be desirable, an
                       offset can be added to all feature values. so
                       that sensible values for entropy are returned.

    Returns:

    A df containing the entropy computed for all times for all columns

    """
    df = df.fillna(0.0)
    if not df.index.is_monotonic:
        df = df.sort_index()

    df_index = df.index

    times, time_to_index, index_to_time = utilities._map_times(df)

    features = utilities._map_features(df)

    pgids, pgid_to_index, index_to_pgid = utilities._map_pgids_1d(df)

    tensor3d = utilities._df_to_tensor_strides(df)

    tensor3d += offset

    sum_over_window = np.zeros_like(tensor3d)

    entropy = np.zeros_like(tensor3d)

    for itime in range(len(times)):
        if itime < window - 1:
            istart = 0
        else:
            istart = itime - window + 1

        sum_over_window[itime, :, :] = np.sum(tensor3d[istart:itime+1], axis=0)

        entropy[itime, :, :] = -np.sum(tensor3d[istart:itime+1]/sum_over_window[itime, :, :] *
                                       np.log2(tensor3d[istart:itime+1]/sum_over_window[itime, :, :]), axis=0)

    df_entropy = entropy_to_df_strides(df, entropy, times, pgids, features, df_index)

    return df_entropy


def entropy_to_df_strides(
        df,
        entropy,
        times,
        pgids,
        features,
        df_index
):
    dim0, dim1, dim2 = len(times), len(pgids), len(features)

    entropy_strides = entropy.strides

    offset2 = entropy_strides[2]

    offset1 = entropy_strides[1]

    offset0 = entropy_strides[0]

    flat = np.lib.stride_tricks.as_strided(entropy, shape=(dim0 * dim1, dim2),
                                           strides=(offset1, offset2))

#    df_column_names = ['entropy_'+ feature for feature in features]

    df_column_names = df.columns

    df_entropy = pd.DataFrame(flat, index=df_index, columns=df_column_names)

    return df_entropy



import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
from views_transformation_library import utilities


def get_spacetime_distances(df, return_values='distances', k=1, nu=1.0, power=0.0):
    """
    
    get_spacetime_distances
    
    For every point in the supplied df, uses scipy.spatial cKDTree to find the nearest k
    past events (where an event is any non-zero value in the input df) and returns either 
    the mean spacetime distance to the k events, or the mean of 
    (size of event)/(spacetime distance)**power.
    
    Arguments:
    
    df:            input df
    
    return_values: choice of what to return. Allowed values:
     
                   distances - return mean spacetime distance to nearest k events     
    
                   weights   - return mean of (size of event)/(spacetime distance)**power
                   
    k:             number of nearest events to be averaged over
    
    nu:            weighting to be applied to time-component of distances, so that 
                   spacetime distance = sqrt(delta_latitude^2 + delta_longitude^2 +
                   nu^2*delta_t^2)
                   
    power:         power to which distance is raised when computing weights. Negative
                   values are automatically converted to positive values.
    
    """

    if return_values not in ['distances', 'weights']:
        raise Exception("unknown return_values; ", return_values,
                        " - allowed choices: distances weights")
                        
    power = np.abs(power)

    times, time_to_index, index_to_time = utilities._map_times(df)

    pgids, pgid_to_longlat, longlat_to_pgid, pgid_to_index, index_to_pgid, ncells, pwr =\
                                                           utilities._map_pgids_2d(df)

    features = utilities._map_features(df)

    use_stride_tricks = True

    tensor = utilities._build_4d_tensor(
                                       df,
                                       pgids,
                                       pgid_to_index,
                                       times,
                                       time_to_index,
                                       ncells,
                                       ncells,
                                       pgid_to_longlat,
                                       features,
                                       use_stride_tricks
                                       )

    df_stdist = space_time_distances(
                                     df,
                                     tensor,
                                     times,
                                     pgids,
                                     features,
                                     longlat_to_pgid,
                                     time_to_index,
                                     ncells,
                                     return_values,
                                     k,
                                     nu,
                                     power
                                     )
                                    
    return df_stdist


def space_time_distances(
                          df,
                          tensor,
                          times,
                          pgids,
                          features,
                          longlat_to_pgid,
                          time_to_index,
                          ncells,
                          return_values,
                          k,
                          nu,
                          power
                          ):
                          
    """
    
    space_time_distances
    
    Finds spacetime distances, scaling spatial distances to degrees (for continuity with
    previous versions of this function) and stretching the time axis by nu.
    
    If k>1, averaging is performed.
    
    """

    PGID_TO_DEGREES = 0.5

    final = np.empty((len(times)*len(pgids), len(features)))

    npg = len(pgids)
    pgids_for_index = []

    for ilong in range(ncells):
        for ilat in range(ncells):
            if (ilong, ilat) in longlat_to_pgid.keys():
                pgid = longlat_to_pgid[(ilong, ilat)]
                pgids_for_index.append(pgid)

    for ifeature, feature in enumerate(df.columns):

        tensor3d = tensor[:, :, :, ifeature]

        for time in times[0:600]:
    
            tindex = time_to_index[time]
            ipgid = 0

            points = np.array(np.where(tensor3d[:, :, :tindex+1] > 0)).astype(float).T

            if len(points) == 0:
                pass
            else:
                points[:, 0] *= PGID_TO_DEGREES
                points[:, 1] *= PGID_TO_DEGREES
                points[:, 2] *= nu
                btree = cKDTree(data=points, leafsize=20)
                
            for ilong in range(ncells):
                for ilat in range(ncells):

                    try:
   
                        pgid = longlat_to_pgid[(ilong, ilat)]
                  
                        if len(points) == 0:

                            feature = 999.
 
                        else:

                            sptime_dists, ipoints = btree.query([ilong*PGID_TO_DEGREES, ilat*PGID_TO_DEGREES,
                                                                 nu*tindex], k)

                            if k == 1:
                                ipoints = [ipoints, ]
                                sptime_dists = [sptime_dists, ]
                                
                            if return_values == 'distances':
                                
                                if k == 1:
                                    feature = sptime_dists[0]
                                else:
                                    feature = np.mean(sptime_dists)
                        
                            elif return_values == 'weights':

                                featurei = np.zeros(k)
                           
                                for i, ipoint in enumerate(ipoints):

                                    coords = points[ipoint]
                                
                                    ilongp = int(coords[0]/PGID_TO_DEGREES)
                                    ilatp = int(coords[1]/PGID_TO_DEGREES)
                                    itimep = int(coords[2]/nu)

                                    if sptime_dists[i] == 0.0:
                                        featurei[i] = tensor[ilongp, ilatp, itimep, 0]
                                    else:
                                        featurei[i] = tensor[ilongp, ilatp, itimep, 0]/(sptime_dists[i]**power)
                        
                                feature = np.mean(featurei)
                
                        indx = tindex*npg+ipgid
                    
                        final[indx, ifeature] = feature

                        ipgid += 1
                    except:
                        pass

    index_names = df.index.names
        
    df_index = pd.MultiIndex.from_product([times, pgids_for_index], names=index_names)
    
#    if return_values=='distances':
#        colnames=['st_distances_nu_'+str(nu)+'_'+feature for feature in features]
#    else:
#        colnames=['st_weights_nu_'+str(nu)+'_k_'+str(k)+'_pwr_'+str(power)+'_'+feature for feature in features]

    colnames = df.columns

    df_stdist = pd.DataFrame(data=final, columns=colnames, index=df_index)

    df_stdist = df_stdist.sort_index()

    return df_stdist



import numpy as np
import pandas as pd
import stepshift
from stepshift import cast
from views_transformation_library import utilities


def get_splag_country(
        df:           pd.DataFrame,
        kernel_inner: int = 1,
        kernel_width: int = 1,
        kernel_power: int = 0,
        norm_kernel:  int = 0)-> pd.DataFrame:

    """
    get_splag_country
    =================

    parameters:
        df (pandas.DataFrame): single-column dataframe containing feature to be
                               transformed.

        kernel_inner (int):    Inner radius of convolution kernel - '1'
                               represents the target country, '2' represents
                               the target country plus its first-order
                               neighbours, and so on.

        kernel_outer (int):    Width of convolution kernel - '1' represents
                               (kernel_inner+1)-th order neighbours,  '2'
                               represents (kernel_inner+2)-th order neighbours,
                               and so on.

        kernel_power (int):    Countries are weighted by:
                               (distance from target country)^kernel_power 
                               kernel_power=0 results in unweighted results

        norm_kernel (int):     If set to 1, the sum of the weights over all
                               neighbours for a given target country is
                               normalised to 1.0

    returns:
        pandas.DataFrame: Dataframe with the lagged variable

    Performs convolutional spatial lags at the country level.

    Country first-order neighbours are obtained directly from the
    country_country_month_ expanded table and represented as a month x country
    x country tensor.

    n-th order neighbours are obtained iteratively.

    The lagged variable is named following this pattern:
    "splag_{kernel_inner}_{kernel_outer}_{kernel_power}_{feature_name}"
    
    """

    month_to_index, country_to_index, splag = splag_cm(
        df, kernel_inner, kernel_width, kernel_power, norm_kernel
    )

    df_splag = splags_to_df(
        df,
        splag,
        kernel_inner,
        kernel_width,
        kernel_power,
        norm_kernel,
        month_to_index,
        country_to_index,
    )

    return df_splag


def get_country_neighbours(
    neighbs,
    ninner,
    nouter,
    month_id,
    month_to_index,
    country_to_index,
    index_to_country,
    neighb_tensor_data,
):

    """

    get_country_neighbours

    For a given target country, an inner radius and an outer radius, fetches list of
    countries between the two radii.

    """

    if ninner < 0:
        inner_neighbs=[]
        ninner=0
    else:
        inner_neighbs = neighbs.copy()
    
    inner_neighbs = get_nth_order_neighbours_from_tensor(
        inner_neighbs,
        month_id,
        ninner,
        month_to_index,
        country_to_index,
        index_to_country,
        neighb_tensor_data,
    )

    if nouter < 0:
        outer_neighbs=[]
        nouter=0
    else:
        outer_neighbs = neighbs.copy()
    
    outer_neighbs = get_nth_order_neighbours_from_tensor(
        outer_neighbs,
        month_id,
        nouter,
        month_to_index,
        country_to_index,
        index_to_country,
        neighb_tensor_data,
    )

    neighbs = np.sort(list(set(outer_neighbs).difference(set(inner_neighbs))))

    return neighbs


def get_nth_order_neighbours_from_tensor(
    neighbs,
    month_id,
    norder,
    month_to_index,
    country_to_index,
    index_to_country,
    neighb_tensor_data,
):

    """

    get_nth_order_neighbours

    Fetches nth order neighbours around target country *including the target country*

    """

    if norder == 0:
        return neighbs
    else:

        neighbscopy = neighbs.copy()

        month_index = month_to_index[month_id]

        for country in neighbs:

            country_index = country_to_index[country]

            neighb_row = neighb_tensor_data[month_index, country_index, :]

            new_neighb_indices = np.where(neighb_row)[0]

            new_neighbs = [index_to_country[n] for n in new_neighb_indices]

            for newn in new_neighbs:

                if newn not in neighbscopy:

                    neighbscopy.append(newn)

        neighbs = neighbscopy

        norder -= 1

        return get_nth_order_neighbours_from_tensor(
            neighbs,
            month_id,
            norder,
            month_to_index,
            country_to_index,
            index_to_country,
            neighb_tensor_data,
        )


def splag_cm(df, kernel_inner, kernel_width, kernel_power, norm_kernel):

    """

    splag_cm

    Uses neighbours and distances tensors to compute, for every country at every month,
    a spatial lag over the requested sets of neighbours.

    Returns a month x country tensor.

    """

    df_matrix = cast.views_format_to_castable(df)

    df_tensor = cast.time_unit_feature_cube(df_matrix)

    data_month_ids = df_tensor.coords["time"].values
    data_country_ids = df_tensor.coords["unit"].values
    data_feature_ids = df_tensor.coords["feature"].values

    # build dicts to transform between month and index and country and index in the
    # data, and in the neighbours tensor

    data_month_to_index = {}
    data_index_to_month = {}
    for imonth, month in enumerate(data_month_ids):
        data_month_to_index[month] = imonth

    data_country_to_index = {}
    data_index_to_country = {}
    for icountry, country in enumerate(data_country_ids):
        data_country_to_index[country] = icountry
        data_index_to_country[icountry] = country

    neighb_tensor = utilities._get_country_neighbours_tensor()

    neighb_month_ids = neighb_tensor.coords["month"].values
    neighb_country_ids = neighb_tensor.coords["country"].values

    neighb_month_to_index = {}
    neighb_index_to_month = {}
    for imonth, month in enumerate(neighb_month_ids):
        neighb_month_to_index[month] = imonth

    neighb_country_to_index = {}
    neighb_index_to_country = {}
    for icountry, country in enumerate(neighb_country_ids):
        neighb_country_to_index[country] = icountry
        neighb_index_to_country[icountry] = country

    distances = utilities._get_country_distances(
        data_country_ids, data_country_to_index
    )

    neighb_tensor_data = neighb_tensor.values

    splag = np.zeros_like(df_tensor.values)

    ninner = kernel_inner - 1
    nouter = ninner + kernel_width 

    for month_id in data_month_ids:
        data_month_index = data_month_to_index[month_id]
        if month_id in neighb_month_ids:

            for country_id in data_country_ids:
                data_country_index = data_country_to_index[country_id]

                if country_id in neighb_country_ids:
                    neighbs = get_country_neighbours(
                        [
                            country_id,
                        ],
                        ninner,
                        nouter,
                        month_id,
                        neighb_month_to_index,
                        neighb_country_to_index,
                        neighb_index_to_country,
                        neighb_tensor_data,
                    )

                    neighbs = [n for n in neighbs if n in data_country_ids]
                else:
                    neighbs = []

                neighbs_data_indices = [data_country_to_index[n] for n in neighbs]

                weights = (
                    distances[data_country_index, neighbs_data_indices] ** kernel_power
                    )

                if norm_kernel:
                    weights /= np.sum(weights)

                for ifeature, feature in enumerate(data_feature_ids):

                    vals = df_tensor[data_month_index, neighbs_data_indices, ifeature].values

                    vals[vals == np.inf] = 0.0

                    splag[data_month_index, data_country_index, ifeature] = np.sum(vals * weights)

#                print('vals',vals)
#                print('weights',weights)
#                print('sum',np.sum(vals * weights))

    return data_month_to_index, data_country_to_index, splag


def splags_to_df(
    df,
    splag,
    kernel_inner,
    kernel_outer,
    kernel_power,
    norm_kernel,
    month_to_index,
    country_to_index,
):

    """

    splags_to_df

    Transforms splag tensor to a dataframe.

    """

    months = np.array(list(idx[0] for idx in df.index.values))

    countries = np.array(list(idx[1] for idx in df.index.values))

    features = df.columns

    flat = np.empty((len(countries), len(features)))

    iflat = 0
    for month, country in zip(months, countries):
        imonth = month_to_index[month]
        icountry = country_to_index[country]

        flat[iflat, :] = splag[imonth, icountry, :]

        iflat += 1

    df_index = df.index

    df_splag = pd.DataFrame(flat, index=df_index, columns=df.columns)

    return df_splag



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['views_transformation_library']

package_data = \
{'': ['*']}

install_requires = \
['ingester3>=0.6.0',
 'pandas>=1.2.3,<2.0.0',
 'scikit_learn>=1.0.2,<2.0.0',
 'scipy>=1.6.2,<2.0.0',
 'stepshift>=1.2.0',
 'xarray>=0.19.0']

setup_kwargs = {
    'name': 'views-transformation-library',
    'version': '2.6.1',
    'description': 'A package containing data transformation functions used by the ViEWS team',
    'long_description': '\n# ViEWS Transformation Library\n\nThis package contains transforms made available in ViEWS cloud.  The transforms\nare all locally available, and do not depend on remote resources or\ncredentials.\n\n## Contributing: \n\nThere is an urgent need to fill this library with transforms.\nTransforms we need, that are not implemented, are listed under issues.\nPlease follow the guidelines for contributing outlined in the [CONTRIBUTING.md](CONTRIBUTING.md) document.\n\n## Usage:\n\n```\npip install views-transformation-library\n```\n\n```\nfrom views_transformation_library.timelag import timelag\nimport viewser\n\nged_best_ns = viewser.get_variable("priogrid_month","priogrid_month.ged_best_ns",2010,2020)\nlagged_data = timelag(ged_best_ns,10)\n\nlagged_from_api = viewser.get_variable("priogrid_month","ged_best_ns",2010,2020,transforms:[\n   {"type":"tlag","args":[10]}\n])\n\nassert lagged_from_api == lagged_data\n```\n\n## API:\n\nEach transform function operates on a pandas dataframe.\n',
    'author': 'peder2911',
    'author_email': 'pglandsverk@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.github.com/prio-data/views_transformation_library',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import numpy as np
import pandas as pd
from views_transformation_library import utilities


def get_tree_lag(df, thetacrit, dfunction_option):
    """
    get_tree_lag

    Driver function for computing tree-lagged features

    Arguments:

    df: dataframe containing one or more features to be lagged

    thetacrit: opening angle used to decide whether to open nodes - large values cause more
    aggressive aggregation of nodes

    dfunction_option: an integer selecting which distance weighting to use:
                
                     - 0: ln(1+d) weighting
                     
                     - 1: 1/d weighting
                     
                     - 2: 1/d^2 weighting  

    """

    df = df.fillna(0.0)
    if not df.index.is_monotonic:
        df = df.sort_index()

    dfunctions = get_dfunctions(dfunction_option)

    tree = SpatialTree()

    tree.build_tree(df)

    tree.stock(df)

    tree.walk(thetacrit, dfunctions)

    df_treelags = tree.tree_lag(df)

    return df_treelags


def get_dfunctions(dfunction_option):

    dfunctions = {}

    if dfunction_option == 0:

        def logd(d):
            return 1./np.log(1.+d)

        dfunctions['logd'] = logd

    elif dfunction_option == 1:

        def d_1(d):
            return 1./d

        dfunctions['d_1'] = d_1

    elif dfunction_option == 2:

        def d_2(d):
            return 1./d/d

        dfunctions['d_2'] = d_2

    else:
        raise Exception(
                "Unrecognised distance function in tree. "
                f"Allowed: 0=1/ln(1+d), 1=1/d, 2=1/d^2, Supplied: {dfunction_option}"
            )

    return dfunctions


def get_grid_lag(df, threshold, dfunctions, split_criterion, keep_grids):

    """
    get_grid_lag

    Driver function for computing adaptive-grid-lagged features

    Arguments:

    df: dataframe containing one or more features to be lagged

    threshold: number to be used to decide whether to split nodes

    dfunctions: a dict of function_name:function pairs, where the functions must take a single
    argument (distance between nodes) and return a single result (weight corresponding to distance)

    split_criterion: list containing one or more of the following strings:

    - 'min': nodes are split if their value of the feature exceeds the value of threshold

    keep_grids: flag instructing function to return the nodeids in the adaptive
    mesh at all timesteps

    """

    tree = SpatialTree()

    tree.build_tree(df)

    tree.stock(df)

    df_gridlags = tree.grid_lag(df, dfunctions, split_criterion, threshold, keep_grids)

    return df_gridlags


class SpatialNode():
    """
    SpatialNode() class
    Defines nodes for insertion in SpatialTree() class

    Nodes have:
    - nodeid: a unique id
    - masterindex, levelindex, level: book-keeping indices giving their tree level and
      location on that level (not needed once tree has been built)
    - parent: parent node
    - left, right, bottom, top: node boundaries
    - centre: node's geometric centre
    - isleaf: a flag indicating whether a leaf node
    - pgid: the corresponding priogrid cell, if a leaf node (-1 if not a leaf node)
    - nleaf: the number of leaf nodes contained within the node's boundaries
    - children: list of node's children
    - features: dict of features loaded into the node by the .stock() method

    """
    def __init__(self):
        self.nodeid = -1
        self.masterindex = -1
        self.levelindex = -1
        self.level = -1
        self.parent = -1
        self.left = -1
        self.right = -1
        self.top = -1
        self.bottom = -1
        self.centre = np.array([-1.0, -1.0])
        self.isleaf = False
        self.pgid = -1
        self.nleaf = 0
        self.children = []
        self.features = {}

    def __repr__(self):
        return f""" nodeid: {self.nodeid}\n masterindex: {self.masterindex}\n levelindex: {self.levelindex}
        level: {self.level}\n parent: {self.parent}\n left: {self.left}
        right: {self.right}\n bottom: {self.bottom}\n top: {self.top}
        centre: {self.centre}\n isleaf: {self.isleaf}\n children: {self.children}"""


class SpatialTree():
    """
    SpatialTree() class
    Defines tree class and methods for building, walking and stocking

    """

    def __init__(self):
        self.ncells = 0
        self.power = 0
        self.pgids = None
        self.pgid_to_longlat = {}
        self.longlat_to_pgid = {}
        self.pgid_to_index = {}
        self.index_to_pgid = {}
        self.nodes = []
        self.interaction_lists = None
        self.interaction_weights = None
        self.times = None
        self.features = None
        self.weightfunctions = None

    def build_tree(self, df):

        """
        build_tree

        This function builds the tree from the leaf nodes downwards.

        A list with space for all possible nodes is first built and populated with the
        leaf nodes actually present in the df the tree was initialised with.

        masterindex is a book-keeping device which is a unique location in the list of
        all possible nodes.

        levelindex is retained for debugging purposes - each node has a number indicating
        its position in its level, constructed in the same way one would number the
        squares on a chessboard, stating with a1, b1, c1, etc.

        Once leaf nodes are populated, work down through the levels, using nodes in the
        level above to generate parent nodes in the current level, using masterindex to
        avoid creating the same parent twice.

        Once this is done, the list of all possible nodes is compactified to discard
        un-needed indices, and all nodeids are reassigned so that the final list of nodes
        is contiguous.

        """

        self.pgids, self.pgid_to_longlat, self.longlat_to_pgid, self.pgid_to_index,\
        self.index_to_pgid, self.ncells, self.power = utilities._map_pgids_2d(df)

        powers = [((2**p)**2) for p in range(self.power+1)]

        cupowers = np.cumsum(powers)

        nodes_provisional = [None for i in range(cupowers[-1])]

        # populate leaf nodes

        for pgid in self.pgids:
            ix, iy = self.pgid_to_longlat[pgid]

            levelindex = ix+iy*self.ncells
            masterindex = cupowers[-2]+levelindex+1

            if nodes_provisional[masterindex] is None:
                node = SpatialNode()
                node.level = self.power
                node.levelindex = levelindex
                node.masterindex = masterindex
                node.left = ix
                node.right = ix+1
                node.bottom = iy
                node.top = iy+1
                node.centre = np.array([ix+0.5, iy+0.5])
                node.isleaf = True
                node.nleaf = 1
                node.pgid = pgid

                nodes_provisional[masterindex] = node

        # populate lower levels

        for p in range(self.power-1, -1, -1):
            istart = cupowers[p]-1
            iend = cupowers[p+1]
            ncellsp = 2**p
            nodesize = self.ncells/ncellsp
            for inode in range(istart, iend):
                if nodes_provisional[inode] is not None:
                    node = nodes_provisional[inode]
                    ix = int(node.centre[0]/float(self.ncells)*float(ncellsp))
                    iy = int(node.centre[1]/float(self.ncells)*float(ncellsp))
                    levelindex = ix+iy*ncellsp
                    if (p-1) < 0:
                        masterindex = 0
                    else:
                        masterindex = cupowers[p-1] + levelindex

                    if nodes_provisional[masterindex] is None:
                        parent = SpatialNode()
                        parent.level = p
                        parent.levelindex = levelindex
                        parent.masterindex = masterindex
                        parent.left = ix*nodesize
                        parent.right = parent.left+nodesize
                        parent.bottom = iy*nodesize
                        parent.top = parent.bottom+nodesize
                        parent.centre = np.array([(parent.left+parent.right)/2, (parent.bottom+parent.top)/2])
                        parent.children.append(node.masterindex)
                        parent.nleaf += node.nleaf
                        nodes_provisional[masterindex] = parent
                    else:
                        parent = nodes_provisional[masterindex]
                        parent.children.append(node.masterindex)
                        parent.nleaf += node.nleaf

                    node.parent = parent.masterindex

        # compactify list of nodes, and reassign unique indices

        provisional_to_final = {}

        for node in nodes_provisional:
            if node is not None:
                self.nodes.append(node)
                node.nodeid = len(self.nodes)-1
                provisional_to_final[node.masterindex] = node.nodeid

        for node in self.nodes:
            if node.parent != -1:
                node.parent = provisional_to_final[node.parent]
            node.children = [provisional_to_final[child] for child in node.children]

    def walk(self, thetacrit, dfunctions):

        """
        walk

        This function generates the list of nodes any given node will import data from,
        when one lags variables using the tree, as well as weights based on the distance
        between nodes' centres.

        The arguments are

        - thetacrit: angle used to decide if a candidate node should be added to a target
          node's interaction list, based on the size of the candidate node and
          the distance between the candidate node and the target node

        - dfunctions: dict of named functions of internode distance d only, used to
          compute weights of node-node interactions
        """

        interaction_lists = {}
        interaction_weights = {}
        self.weightfunctions = dfunctions

        for targetnode in self.nodes:
            if not targetnode.isleaf:
                continue
            targetpos = targetnode.centre
            targetid = targetnode.pgid
            interaction_list = []
            interaction_weight = {}
            nodestodo = [0, ]

            while len(nodestodo) > 0:
                inode = nodestodo[0]
                node = self.nodes[inode]

                d = targetpos - node.centre + 0.01
                d = np.sqrt(np.dot(d, d))
                h = node.top - node.bottom
                theta = h/d
                if theta > thetacrit:
                    # attempt to split node
                    if node.isleaf:
                        # if node is a leaf node, it cannot be split: add to interaction list
                        if not(targetnode.nodeid == node.nodeid):
                            interaction_list.append(node.nodeid)
                            weights = {}
                            for dfunction in dfunctions.keys():
                                weights[dfunction] = dfunctions[dfunction](d)
                            interaction_weight[node.nodeid] = weights

                        nodestodo.pop(0)
                    else:
                        # otherwise get children and put in list of nodes to check
                        for ichild in node.children:
                            nodestodo.append(ichild)

                        nodestodo.pop(0)
                else:
                    # add to interaction list
                    if node.nleaf > 0:
                        # TODO not sure this check is still necessary
                        interaction_list.append(node.nodeid)
                        weights = {}
                        for dfunction in dfunctions.keys():
                            weights[dfunction] = dfunctions[dfunction](d)
                        interaction_weight[node.nodeid] = weights

                    nodestodo.pop(0)

            interaction_lists[targetid] = interaction_list
            interaction_weights[targetid] = interaction_weight

        self.interaction_lists = interaction_lists
        self.interaction_weights = interaction_weights

        return

    def stock(self, df, use_stride_tricks=True):

        """
        stock

        This function loads features into the tree.

        Data from the input df is first cast to a 3D tensor with dims time x pgid x
        features.

        The tensor is used to populate the leaf nodes, and values then just propagate
        down through the tree via the parents.

        At present, values of parents are the *sums* of those of their children.

        """

        self.features = utilities._map_features(df)

        self.times, _, _ = utilities._map_times(df)

        for node in self.nodes:
            for feature in self.features:
                node.features[feature] = np.zeros(len(self.times))

                # TODO weighted centres

        if use_stride_tricks:
            tensor3d = utilities._df_to_tensor_strides(df)
        else:
            tensor3d = utilities._df_to_tensor_no_strides(df)

        for ifeature, feature in enumerate(self.features):
            for node in self.nodes:
                if node.isleaf:
                    pgid = node.pgid
                    ipgid = self.pgid_to_index[pgid]
                    vals = tensor3d[:, ipgid, ifeature]

                    node.features[feature] += vals
                    while node.parent >= 0:
                        node = self.nodes[node.parent]

                        node.features[feature] += vals

        return

    def tree_lag(self, df):

        """
        tree_lag

        This function computes features spatially lagged using the tree for every pgid.

        The interaction lists and weights, and the features loaded into each node are
        combined in weighted sums.

        The lagged features are then packed into a dataframe.

        """

        dim0 = len(self.times)
        dim1 = len(self.pgids)
        dim2 = len(self.features)
        nweightfunctions = len(self.weightfunctions)
        weightkeys = list((self.weightfunctions.keys()))

        treelags = np.empty((dim0, dim1, dim2*nweightfunctions))

        for ifeature, feature in enumerate(self.features):
            for node in self.nodes:
                if node.isleaf:
                    pgid = node.pgid
                    ipgid = self.pgid_to_index[pgid]

                    interactions = self.interaction_lists[pgid]
                    for iweight, weightkey in enumerate(weightkeys):
                        weights = np.array([self.interaction_weights[pgid][partner][weightkey] for partner in
                                            interactions])

                        sums = np.zeros(len(self.times))

                        for partner, weight in zip(interactions, weights):
                            node = self.nodes[partner]
                            sums += node.features[feature]*weight
                        treelags[:, ipgid, ifeature*nweightfunctions+iweight] = sums

        # create tensor to pack into df

        flat = np.empty((dim0*dim1, dim2*nweightfunctions))

        for ichunk in range(dim0):
            flat[ichunk*dim1:(ichunk+1)*dim1, :] = treelags[ichunk, :, :]

        df_column_names = df.columns

        index_names = df.index.names
        
        df_index = pd.MultiIndex.from_product([self.times, self.pgids], names=index_names)

        df_treelags = pd.DataFrame(flat, index=df_index, columns=df_column_names)

        return df_treelags

    def grid_lag(self, df, dfunctions, split_criterion='min', threshold=10000, keep_grids=False):

        """
        grid_lag

        This function uses the tree in a different way to build, at each timestep, a single
        grid from the tree nodes, with nodes being split according to a refinement
        criterion. The available criteria are:

        - 'min': starting with the root node, nodes whose value at a given timestep is
          greater than the threshold are recursively split

        """

        if keep_grids:
            grids=[]

        dim0 = len(self.times)
        dim1 = len(self.pgids)
        dim2 = len(self.features)

        self.weightfunctions = dfunctions

        nweightfunctions = len(self.weightfunctions)

        weightkeys = list((self.weightfunctions.keys()))

        gridlags = np.empty((dim0, dim1, dim2*nweightfunctions))

        def split_min(threshold, time, feature):
            grid_nodes = []
            nodestodo = [0, ]

            while len(nodestodo) > 0:

                inode = nodestodo.pop(0)
                node = self.nodes[inode]
                val = node.features[feature][time]

                if val > threshold:
                    if node.isleaf:

                        grid_nodes.append(inode)
                    else:
                        for child in node.children:

                            nodestodo.append(child)

                else:
                    if node.nleaf > 0:

                        grid_nodes.append(inode)

            return grid_nodes

        if split_criterion == 'min':
            split_decider = split_min

        else:
            print('Unknown criterion for splitting nodes: ', split_criterion)
            raise Exception()

        for itime, time in enumerate(self.times):
            for ifeature, feature in enumerate(self.features):

                grid_nodes = split_decider(threshold, itime, feature)

                if keep_grids:
                    grids.append(grid_nodes)

                for targetnode in self.nodes:
                    if not targetnode.isleaf:
                        continue
                    targetpos = targetnode.centre

                    targetnodeid = targetnode.nodeid

                    sums = np.zeros((nweightfunctions))

                    for inode in grid_nodes:
                        node = self.nodes[inode]

                        if node.nodeid == targetnodeid:
                            continue

                        val = node.features[feature][itime]

#                        d=np.array(targetpos)-np.array(node.centre)+0.01
                        d = targetpos - node.centre
                        d = np.sqrt(np.dot(d, d))

                        for iweight, weightkey in enumerate(weightkeys):
                            sums[iweight] += val*dfunctions[weightkey](d)

                    ipgid = self.pgid_to_index[targetnode.pgid]
                    gridlags[itime, ipgid, ifeature*nweightfunctions:(ifeature+1)*nweightfunctions] = sums

        flat=np.empty((dim0*dim1,dim2*nweightfunctions))

        for ichunk in range(dim0):
            flat[ichunk*dim1:(ichunk+1)*dim1, :] = gridlags[ichunk, :, :]

        df_column_names = df.columns

        index_names = df.index.names
        
        df_index = pd.MultiIndex.from_product([self.times, self.pgids], names=index_names)
        
        df_gridlags = pd.DataFrame(flat, index=df_index, columns=df_column_names)

        if keep_grids:
            return df_gridlags, grids
        else:
            return df_gridlags








import numpy as np
import pandas as pd


def add_column(df):

    df_left, df_right = get_left_right(df)

    df_sum = pd.DataFrame(index=df_left.index)

    df_sum[df_left.columns[0]] = df_left.values + df_right.values

    return df_sum


def subtract_column(df):

    df_left, df_right = get_left_right(df)

    df_diff = pd.DataFrame(index=df_left.index)

    df_diff[df_left.columns[0]] = df_left.values - df_right.values

    return df_diff


def multiply_column(df):

    df_left, df_right = get_left_right(df)

    df_mult = pd.DataFrame(index=df_left.index)

    df_mult[df_left.columns[0]] = df_left.values * df_right.values

    return df_mult


def divide_column(df):

    df_left, df_right = get_left_right(df)

    df_div = pd.DataFrame(index=df_left.index)

    df_div[df_left.columns[0]] = df_left.values / df_right.values

    return df_div


def get_left_right(df):

    assert len(df.columns) == 2

    left_column = df.columns[0]

    right_column = df.columns[1]

    df_left = pd.DataFrame(df[left_column])

    df_right = pd.DataFrame(df[right_column])

    return df_left, df_right
